(ns fluree.db.enterprise
  (:require [fluree.crypto :as crypto]
            [alphabase.core :as alphabase]
            [clojure.string :as str])
  (:import (java.nio ByteBuffer)))

;; Version timestamp used to validate license key timestamp - update with each release.
;; A license key is invalid if the encoded timestamp is < version timestamp.
;; Generate a new timestamp with the release date using:
;; (.toEpochMilli (java.time.Instant/parse "2019-12-22T00:00:00Z"))
(def ^:const version-timestamp 1568073600000)               ;; 1568073600000 = "2019-09-10T00:00:00Z"


;; public keys Fluree uses, signature will be validated to ensure public key matches
(def ^:const fluree-public-keys #{"0297a2f24fdc9af30f0dbabc1d7e4f36dc8b20e231878d8032bd1d839af54b245a"})


(defn generate-license-key
  "Generates a new license key, returns byte array."
  [server-n valid-until-epoch license-id]
  (assert (<= server-n (Short/MAX_VALUE)))
  (assert (<= valid-until-epoch (Long/MAX_VALUE)))
  (assert (<= license-id (Long/MAX_VALUE)))
  (let [key-version 1                                       ;; must be <= 32767
        bb          (ByteBuffer/allocate 32)]
    (doto bb
      (.putShort key-version)
      (.putShort server-n)
      (.putLong valid-until-epoch)
      (.putLong license-id))
    (.array bb)))


(defn formatted-license-key
  [server-n valid-until-epoch license-id]
  (let [private-key        "144decd46f9c5f78b7f1845e26e7938812df056f4b63c2e960861afb85a88dca"
        license-key        (generate-license-key server-n valid-until-epoch license-id)
        sig                (crypto/sign-message license-key private-key)
        license-key-base58 (alphabase/bytes->base58 license-key)
        sig-base58         (alphabase/base-to-base sig :hex :base58)]
    (str license-key-base58 "." sig-base58)))


(defn- decode-license-key
  "Decodes license key as a byte-array and returns a map with license values."
  [key-ba]
  (assert (bytes? key-ba))
  (try
    (let [bb                (ByteBuffer/wrap key-ba)
          version           (.getShort bb)
          server-n          (.getShort bb)
          valid-until-epoch (.getLong bb)
          license-id        (.getLong bb)]
      {:v     version
       :n     server-n
       :until valid-until-epoch
       :id    license-id})
    (catch Exception e
      (throw (ex-info (str "Invalid license key provided: " (alphabase/bytes->base58 key-ba))
                      {:status 500
                       :error  :db/invalid-license} e)))))


(defn decode-formatted-key
  "Main method for use with Fluree. Will return map of license data, but only
  if signature provided is valid."
  [base-58-key-with-sig]
  (let [[license-key sig] (str/split base-58-key-with-sig #"\.")
        license-key-ba (alphabase/base58->bytes license-key)
        sig-hex        (alphabase/base-to-base sig :base58 :hex)
        license-data   (decode-license-key license-key-ba)
        pub-key        (try (crypto/pub-key-from-message license-key-ba sig-hex)
                            (catch Exception _ nil))
        valid-pub-key? (contains? fluree-public-keys pub-key)]
    (when-not valid-pub-key?
      (throw (ex-info (str "Invalid signature for license key.")
                      {:status 500
                       :error  :db/invalid-license})))
    license-data))




(comment

  (def lkey (let [n-server    3
                  valid-until (.toEpochMilli (java.time.Instant/parse "2020-12-22T00:00:00Z"))
                  license-id  1234]
              (formatted-license-key n-server valid-until license-id)))

  (decode-formatted-key lkey)

  )