(ns fluree.db.ledger.snapshot
  (:require [clojure.java.io :as io]
            [abracad.avro :as avro]
            [fluree.db.serde.avro :as avro-serde]
            [fluree.db.storage.core :as storage]
            [fluree.db.util.async :refer [<? <?? go-try]]
            [clojure.core.async :as async]
            [fluree.db.session :as session]
            [fluree.db.ledger.txgroup.txgroup-proto :as txproto]
            [fluree.db.ledger.storage.filestore :as filestore]
            [fluree.db.ledger.consensus.dbsync2 :as dbsync2]
            [fluree.db.ledger.reindex :as reindex]
            [fluree.crypto :as crypto]))

(defn snapshot-file-path
  [network dbid]
  (str network "/" dbid "/snapshot/" (System/currentTimeMillis) ".avro"))


(defn create-snapshot
  "Puts all blocks into a single log, AVRO encoded, and zipped.

  Stores on the local file system."
  ([conn network dbid] (create-snapshot conn network dbid (snapshot-file-path network dbid)))
  ([conn network dbid file-key]
   (go-try
     (let [base-path (or (-> conn :meta :storage-directory) "data")
           file-path (str base-path file-key)]

       ;; ensure path to file exists

       (io/make-parents (io/file file-path))
       ;; ensure there is block data for the DB, else we'd possibly overwrite an existing snapshot with null data
       (when-not (<? (storage/read-block conn network dbid 1))
         (throw (ex-info (str "Snapshot called for db: " network "/" dbid ", however there is no block data for it.")
                         {:status 400
                          :error  :db/invalid-snapshot})))
       (with-open [adf (avro/data-file-writer "snappy" avro-serde/FdbBlock-schema file-path)]
         (loop [block 1]
           (let [block-data (<? (storage/read-block conn network dbid block))]
             (when-not (nil? block-data)
               (.append adf block-data)
               (recur (inc block))))))
       file-key))))

;; TODO - allow snapshots to be located elsewhere, i.e. S3 buckets.
(defn get-snapshot-remote
  [conn storage-dir file-path]
  (if (filestore/exists? (str storage-dir file-path))
    (go-try true)
    (let [remote-copy-fn (dbsync2/remote-copy-fn* conn (-> conn :group :raft :other-servers) 3000)
          res-chan       (async/chan)]
      (remote-copy-fn file-path res-chan))))

(defn read-snapshot
  "Reads a sequence of blocks from an existing snapshot file."
  [file-path]
  (with-open [adf (avro/data-file-reader file-path)]
    (doall (seq adf))))


(defn create-db-from-snapshot
  [conn network dbid snapshot max-block command]
  (go-try
    (let [storage-dir      (-> conn :meta :storage-directory)
          snapshot-exists? (<?? (get-snapshot-remote conn storage-dir snapshot))
          max-block        (if (int? max-block) max-block Integer/MAX_VALUE)
          db-blocks        (read-snapshot (str storage-dir snapshot))
          highest-block    (apply max (map :block db-blocks))
          to-block         (min highest-block max-block)]
      (doseq [block db-blocks]
        (if (<= (:block block) max-block)
          (<? (storage/write-block conn network dbid block))))

      (<? (txproto/register-genesis-block-async (:group conn) network dbid to-block))
      (let [indexed-db (<?? (reindex/reindex conn network dbid))
            txid       (crypto/sha3-256 (:cmd command))
            group      (-> indexed-db :conn :group)
            block      (:block indexed-db)
            _          (<?? (txproto/initialized-ledger-async group txid network dbid block nil (get-in indexed-db [:stats :indexed])))
            sess       (session/session conn (str network "/" dbid))
            _          (session/close sess)]
        indexed-db))))

