(ns fluree.db.query.analytical
  (:require [clojure.set :as set]
            [fluree.db.query.range :as query-range]
            #?(:clj  [clojure.core.async :as async]
               :cljs [cljs.core.async :as async])
            [fluree.db.time-travel :as time-travel]
            [fluree.db.util.async :refer [<? go-try merge-into?]]
            [fluree.db.util.core :as util]
            [fluree.db.query.analytical-wikidata :as wikidata]
            [fluree.db.query.analytical-filter :as filter]
            [clojure.string :as str]
            [fluree.db.util.log :as log]
            #?(:cljs [cljs.reader])))

(defn variable? [form]
  (when (and (or (string? form) (symbol? form)) (= (first (name form)) \?))
    (symbol form)))


(defn clause->rel
  [interm-vars clause]
  (reduce-kv (fn [acc idx key]
               (let [key-as-var   (variable? key)
                     static-value (get interm-vars key-as-var)]
                 (cond static-value
                       (update acc :search #(conj % static-value))

                       key-as-var
                       (-> acc
                           (update :search #(conj % nil))
                           (assoc-in [:rel key-as-var] idx))

                       :else
                       (update acc :search #(conj % key))))) {:search [] :rel {}} clause))


(defn get-ns-arrays [ns arrays]
  (map (fn [array] (map #(nth array %) ns)) arrays))

(defn clause->keys
  [clause]
  (reduce (fn [acc var]
            (if-let [var (variable? var)]
              (conj acc var) acc))
          [] clause))

(defn intersecting-keys-tuples-clause
  [tuples clause]
  (let [rel-keys    (-> tuples :headers set)
        clause-keys (clause->keys clause)]
    (reduce (fn [acc key]
              (if (rel-keys key)
                (conj acc key) acc)) [] clause-keys)))

(defn intersecting-keys-tuples
  [a-tuples b-tuples]
  (let [a-keys (-> a-tuples :headers set)
        b-keys (-> b-tuples :headers)]
    (reduce (fn [acc key]
              (if (a-keys key)
                (conj acc key) acc)) [] b-keys)))

(defn select-from-tuples
  [vars tuples]
  (let [ns (reduce (fn [acc var-smt]
                     (if-let [var (or (variable? var-smt)
                                      (:variable var-smt))]
                       (conj acc (util/index-of (:headers tuples) var))
                       (throw (ex-info (str var-smt " cannot be retrieved from the results. Check that it is declared in your where clause.") {:status 400 :error :db/invalid-query})))) [] vars)]
    (get-ns-arrays ns (:tuples tuples))))

(defn add-fuel
  [add-amount fuel max-fuel]
  (if (and max-fuel (> add-amount max-fuel))
    (throw (ex-info (str "Maximum query fuel exceeded: " max-fuel)
                    {:status 400 :error :db/exceeded-cost}))
    (when (and fuel (volatile? fuel))
      (vswap! fuel + add-amount))))

(defn replace-vars-wikidata
  [all-wd intersecting-vars vars]
  (mapv (fn [clause] (mapv (fn [clause-item]
                             (if-let [key-replace (intersecting-vars (symbol clause-item))]
                               (let [replacement  (get vars key-replace)
                                     replacement' (if (number? (#?(:clj read-string :cljs cljs.reader/read-string) replacement)) replacement (str "\"" replacement "\""))] replacement') clause-item)) clause)) all-wd))

(defn wikidata->tuples
  [q-map clause r {:keys [vars] :as res} optional? fuel max-fuel]
  (go-try
    (if (nil? r)
      (let [all-wd (wikidata/get-all-wd-clauses (:where q-map))]
        ;; If there is a WD clause in the where clause, then we will evaluate
        ;; all the optional WD clauses at the same time as all the other WD clauses.
        ;; therefore, when it comes time to evaluate an optional WD clause, we ignore it.
        (if (and optional? (not (empty? all-wd)))
          [nil r]
          (let [optional          (wikidata/get-all-wd-clauses (:optional q-map))
                all-wd-vars       (-> (apply concat (map clause->keys all-wd)) set)
                all-vars          (into all-wd-vars
                                        (apply concat (map clause->keys optional)))
                intersecting-vars (-> (remove nil? (map #(all-vars %) (keys vars))) set)
                matching-vars     (apply concat (map #(intersecting-keys-tuples-clause res %) all-wd))
                matching-vals     (select-from-tuples matching-vars res)
                all-wd-w-intm     (replace-vars-wikidata all-wd intersecting-vars vars)
                all-vars'         (remove intersecting-vars all-vars)
                wikidataTuples    (<? (wikidata/get-wikidata-tuples q-map all-wd-w-intm matching-vars matching-vals all-vars' optional))
                _                 (add-fuel (-> wikidataTuples :tuples count) fuel max-fuel)]
            [wikidataTuples r])))
      ;; If we are not evaluating the last clause, we drop ALL wikidata clauses from rest.
      ;; Then we add the current WD clause to the end of rest.
      ;; When we finally evaluate WD, we pull from the FULL query map
      [nil (conj (wikidata/drop-all-wd-clauses r) clause)])))



(defn get-source-clause
  [db clause]
  (go-try (let [source (first clause)]
            (cond (= "$fdb" source)
                  [db (into [] (subvec clause 1 4))]

                  ;; block
                  (str/starts-with? source "$fdb")
                  (let [source-part (subs source 4)
                        block       (try (#?(:clj read-string :cljs cljs.reader/read-string) source-part)
                                         (catch #?(:clj Exception :cljs :default) e source-part))
                        block       (if (int? block) block source-part)
                        db          (<? (time-travel/as-of-block db block))
                        _           (when (util/exception? db)
                                      (throw db))]
                    [db (into [] (subvec clause 1 4))])

                  (= "$wd" source)
                  ["$wd" clause]

                  ;; TODO - support cross-db queries
                  ;; other db

                  :else
                  (throw (ex-info (str "The data source: " source " is not supported in Fluree")
                                  {:status 400
                                   :error  :db/invalid-query}))))))

(defn fdb-clause->tuples
  [db {:keys [vars] :as res} clause fuel max-fuel]
  (go-try (let [{:keys [search rel]} (clause->rel vars clause)
                res    (<? (query-range/search db search))
                _      (add-fuel (count res) fuel max-fuel)
                tuples (get-ns-arrays (vals rel) res)]
            {:headers (keys rel)
             :vars    vars
             :tuples  tuples})))

;; Uses SPARQL aggregates + additional ones as extension.
;; https://docs.data.world/tutorials/sparql/list-of-sparql-aggregate-functions.html
(def built-in-aggregates
  (letfn [(sum [coll] (reduce + 0 coll))
          (avg [coll] (/ (sum coll) (count coll)))
          (median
            [coll]
            (let [terms (sort coll)
                  size  (count coll)
                  med   (bit-shift-right size 1)]
              (cond-> (nth terms med)
                      (even? size)
                      (-> (+ (nth terms (dec med)))
                          (/ 2)))))
          (variance
            [coll]
            (let [mean (avg coll)
                  sum  (sum (for [x coll
                                  :let [delta (- x mean)]]
                              (* delta delta)))]
              (/ sum (count coll))))
          (stddev
            [coll]
            (Math/sqrt (variance coll)))]
    {'avg            avg
     'median         median
     'variance       variance
     'stddev         stddev
     'min            (fn
                       ([coll] (reduce (fn [acc x]
                                         (if (neg? (compare x acc))
                                           x acc))
                                       (first coll) (next coll)))
                       ([n coll]
                        (vec
                          (reduce (fn [acc x]
                                    (cond
                                      (< (count acc) n)
                                      (sort compare (conj acc x))
                                      (neg? (compare x (last acc)))
                                      (sort compare (conj (butlast acc) x))
                                      :else acc))
                                  [] coll))))
     'max            (fn
                       ([coll] (reduce (fn [acc x]
                                         (if (pos? (compare x acc))
                                           x acc))
                                       (first coll) (next coll)))
                       ([n coll]
                        (vec
                          (reduce (fn [acc x]
                                    (cond
                                      (< (count acc) n)
                                      (sort compare (conj acc x))
                                      (pos? (compare x (first acc)))
                                      (sort compare (conj (next acc) x))
                                      :else acc))
                                  [] coll))))
     'sum            sum
     'rand           (fn
                       ([coll] (rand-nth coll))
                       ([n coll] (vec (repeatedly n #(rand-nth coll)))))
     'sample         (fn [n coll]
                       (vec (take n (shuffle coll))))
     'count          count
     'groupconcat    concat
     'count-distinct (fn [coll] (count (distinct coll)))}))


(defn aggregate? [x] (and (string? x)
                          (re-matches #"^\(.+\)$" x)))

(defn parse-aggregate [x valid-var]
  (let [list-agg   (#?(:clj read-string :cljs cljs.reader/read-string) x)
        list-count (count list-agg)
        as?        (= 'as (first list-agg))
        func-list  (if as? (second list-agg)
                           list-agg)
        _          (when-not (or (= 3 list-count) (= 2 list-count))
                     (throw (ex-info (str "Invalid aggregate selection, provided: " x)
                                     {:status 400 :error :db/invalid-query})))
        _          (when-not (get built-in-aggregates (first func-list))
                     (throw (ex-info (str "Invalid aggregate selection function, provided: " x)
                                     {:status 400 :error :db/invalid-query})))
        _          (when-not (valid-var (last func-list))
                     (throw (ex-info (str "Invalid select variable in aggregate select, provided: " x)
                                     {:status 400 :error :db/invalid-query})))]
    {:variable (last func-list)
     :as       (if as? (last list-agg) list-agg)
     :code     func-list}))


(defn calculate-aggregate
  [res agg]
  (let [agg-read     (:code agg)
        as           (:as agg)
        agg-function (get built-in-aggregates (first agg-read))
        agg-params   (mapv (fn [param]
                             (if (symbol? param)
                               (into [] (flatten (select-from-tuples [param] res)))
                               param)) (rest agg-read))
        agg-result   (apply agg-function agg-params)]
    [as agg-result]))

(defn add-aggregate-cols
  [res aggregate]
  (reduce (fn [res agg]
            (let [[as agg-result] (calculate-aggregate res agg)
                  {:keys [headers tuples]} res
                  tuples'  (map #(conj % agg-result) tuples)
                  headers' (conj headers as)]
              {:headers headers' :tuples tuples'}))
          res aggregate))

(defn clause->tuples
  "Tuples and optional? are only used for Wikidata, because need to both limit calls to Wikidata,
  and ensure that returned results are as limited as possible (but still relevant)."
  [db q-map res clause r optional? fuel max-fuel]
  (go-try
    (cond (= 3 (count clause))
          [(<? (fdb-clause->tuples db res clause fuel max-fuel)) r]

          (= 2 (count clause))
          (let [[k v] clause
                k             (variable? k)
                _             (when-not k (throw (ex-info (str "Invalid intermediate aggregate value. Provided: " clause)
                                                          {:status 400 :error :db/invalid-query})))
                _             (when-not (aggregate? v) (throw (ex-info (str "Invalid aggregate code. Provided: " clause)
                                                                       {:status 400 :error :db/invalid-query})))
                {:keys [headers vars]} res
                validated-agg (parse-aggregate v (set (conj headers vars)))
                [_ agg-result] (calculate-aggregate res validated-agg)]
            [(update res :vars merge {k agg-result}) r])

          :else
          (let [[db clause] (<? (get-source-clause db clause))]
            (if (= "$wd" db)
              (<? (wikidata->tuples q-map clause r res optional? fuel max-fuel))
              [(<? (fdb-clause->tuples db res clause fuel max-fuel)) r])))))


(defn find-match
  "Given a single tuple from A, a-idxs, b-idxs, b-not-idxs, and b-tuples, return any tuples in b that match."
  [a-tuple a-idxs b-tuples b-idxs b-not-idxs]
  (let [a-tuple-part (map #(nth a-tuple %) a-idxs)]
    (reduce (fn [acc b-tuple]
              (if (= a-tuple-part (map #(nth b-tuple %) b-idxs))
                (conj (or acc []) (concat a-tuple (map #(nth b-tuple %) b-not-idxs))) acc))
            nil b-tuples)))

(defn inner-join
  [a-res b-res]
  (let [common-keys (intersecting-keys-tuples a-res b-res)
        a-idxs      (map #(util/index-of (:headers a-res) %) common-keys)
        b-idxs      (map #(util/index-of (:headers b-res) %) common-keys)
        b-not-idxs  (-> b-res :headers count (#(range 0 %))
                        set (set/difference (set b-idxs)) (#(apply vector %)))
        c-tuples    (apply concat (map (fn [a-tuple]
                                         (find-match a-tuple a-idxs (:tuples b-res)
                                                     b-idxs b-not-idxs)) (:tuples a-res)))
        c-headers   (concat (:headers a-res) (map #(nth (:headers b-res) %) b-not-idxs))]
    {:headers c-headers
     :vars    (merge (:vars a-res) (:var b-res))
     :tuples  c-tuples}))

(defn symbolize-keys
  [q-map]
  (let [keys (map symbol (keys q-map))
        vals (vals q-map)]
    (zipmap keys vals)))

(defn where->inner-joins
  [db where {:keys [vars] :as q-map} fuel max-fuel]
  (go-try (loop [[clause & r] where
                 res {}]
            (if clause
              (let [res' (if (empty? res) {:vars (symbolize-keys vars)} res)
                    [next-res r] (<? (clause->tuples db q-map res' clause r false fuel max-fuel))]
                (cond (= 2 (count clause))
                      (recur r next-res)

                      (empty? res)
                      (recur r (or next-res res))

                      (nil? next-res)
                      (recur r res)

                      :else
                      (recur r (inner-join res next-res))))
              res))))

(defn left-outer-join
  "OPTIONAL clause is equivalent to a left outer join. If there are no matches in the b-tuples,
  we just return a 'match' where each element of the match from b-tuple is nil."
  [a-tuples b-tuples]
  (let [common-keys (intersecting-keys-tuples a-tuples b-tuples)
        a-idxs      (map #(util/index-of (:headers a-tuples) %) common-keys)
        b-idxs      (map #(util/index-of (:headers b-tuples) %) common-keys)
        b-not-idxs  (-> b-tuples :headers count (#(range 0 %))
                        set (set/difference (set b-idxs)) (#(apply vector %)))
        c-tuples    (apply concat (map (fn [a-tuple]
                                         (let [matches (find-match a-tuple a-idxs (:tuples b-tuples)
                                                                   b-idxs b-not-idxs)]

                                           (or matches [(concat a-tuple (repeat (count common-keys) nil))])))
                                       (:tuples a-tuples)))
        c-headers   (concat (:headers a-tuples) (map #(nth (:headers b-tuples) %) b-not-idxs))]
    {:headers c-headers
     :vars    (merge (:vars a-tuples) (:vars b-tuples))
     :tuples  c-tuples}))

(defn optional->left-outer-joins
  [db q-map optional-clauses where-tuples fuel max-fuel]
  (async/go-loop [[opt-clause & r] optional-clauses
                  tuples where-tuples]
    (if opt-clause
      (let [[next-tuples r] (<? (clause->tuples db q-map tuples opt-clause r true fuel max-fuel))]
        (if (nil? next-tuples)
          (recur r tuples)
          (recur r (left-outer-join tuples next-tuples))))
      tuples)))


(defn tuples->filter-required
  [headers tuples valid-vars filter-code-req]
  (let [filter-code-req-str  (str "(and " (str/join " " filter-code-req) ")")
        [filter-code-req* _] (or (filter/valid-filter? filter-code-req-str valid-vars)
                                 (throw (ex-info (str "Invalid required filters, provided: " filter-code-req-str)
                                                 {:status 400 :error :db/invalid-query})))
        filter-code-req-str* (str filter-code-req*)]
    (filter #(filter/filter-row headers % filter-code-req-str*) tuples)))

(defn tuples->filter-optional
  [headers tuples valid-vars filter-code-opts]
  (reduce (fn [tuples filt]
            (let [var-atom        (atom #{})
                  [filt* filt-vars] (or (filter/valid-filter? filt valid-vars var-atom)
                                        (throw (ex-info (str "Invalid filter, provided: " filt)
                                                        {:status 400 :error :db/invalid-query})))
                  filt-str        (str filt*)
                  filt-vars-idxs  (map #(util/index-of headers %) @filt-vars)
                  filtered-tuples (reduce (fn [acc clause]
                                            (if (every? #(nth clause %) filt-vars-idxs)
                                              (if (filter/filter-row headers clause filt-str)
                                                (conj acc clause)
                                                acc)
                                              (conj acc clause)))
                                          [] tuples)]
              filtered-tuples)) tuples filter-code-opts))

(defn tuples->filtered
  [{:keys [headers vars tuples] :as tuple-map} filters]
  (let [valid-vars      (set headers)
        header-vec      (into [] headers)
        _               (when (string? filters)
                          (throw (ex-info (str "Filter must be enclosed in square brackets. Provided: " filters)
                                          {:status 400
                                           :error  :db/invalid-query})))
        filter-code-req (filter/get-filters filters false)
        tuples          (if (not-empty filter-code-req)
                          (tuples->filter-required header-vec tuples valid-vars filter-code-req)
                          tuples)
        filter-code-opt (filter/get-filters filters true)
        tuples          (if (not-empty filter-code-opt)
                          (tuples->filter-optional header-vec tuples valid-vars filter-code-opt)
                          tuples)]
    {:headers headers
     :vars    vars
     :tuples  tuples}))

(defn q
  [q-map fuel max-fuel db]
  (go-try (let [{:keys [where optional filter]} q-map
                where-res    (<? (where->inner-joins db where q-map fuel max-fuel))
                optional-res (if optional
                               (<? (optional->left-outer-joins db q-map optional where-res fuel max-fuel)) where-res)
                filter-res   (if filter
                               (tuples->filtered optional-res filter)
                               optional-res)
                res          filter-res]
            res)))


(comment
  (def conn (:conn user/system))
  (def db (async/<!! (fluree.db.api/db conn "fluree/test")))



  (async/<!! (q {:select ["?chat", "?comment"]
                 :where  [["?chat" "_predicate/name" "?comment"]]} (volatile! 0) 1000 db))

  (async/<!! (fdb-clause->tuples db ['?chat "chat/comments" '?comment] (volatile! 0) 1000000))

  (async/<!! (where->inner-joins db [["?t", "_block/number", "?number"],
                                     ["?maxBlock" "(max ?number)"],
                                     ["?t", "_block/hash", "?hash"]]
                                 {:select   ["?hash", "?pHash", "?number"]
                                  :where    [["?t", "_block/number", "?number"],
                                             {"?maxBlock" "(max ?number)"},
                                             ["?t", "_block/hash", "?hash"]]
                                  :optional [["?t" "_block/prevHash" "?pHash"]]} nil nil))




  (async/<!! (q {:select   ["?handle" "?num"]
                 :where    [["?person" "person/handle" "?handle"]]
                 :optional [["?person" "person/favNums" "?num"]]
                 :filter   [["optional" "(> 10 ?num)"]]} (volatile! 0) 1000 db))

  )