(ns fluree.db.query.http-signatures
  (:require [fluree.crypto :as crypto]
            [clojure.string :as str]
            [fluree.db.util.core :as util]
            [fluree.db.util.log :as log]
            [fluree.db.util.json :as json])
  (:import (java.time ZoneOffset ZonedDateTime)
           (java.time.format DateTimeFormatter)
           (java.net URL)))

;; signatures of http requests. see:
;; https://tools.ietf.org/id/draft-cavage-http-signatures-08.html

(defn generate-signing-string
  "keys must include:
  method - post, get, etc.
  host - host/server name
  path - path of the request
  "
  [val-map ks]
  (let [get-or-throw (fn [m k]
                       (or (get m k)
                           (throw (ex-info (str "Signing string component " k " is not present.")
                                           {:status 401 :error :db/invalid-auth}))))]
    (->> (reduce
           (fn [acc k]
             (if (= "(request-target)" k)
               (conj acc (str "(request-target): "
                              (str/lower-case (name (get-or-throw val-map "method")))
                              " " (str/lower-case (get-or-throw val-map "path"))))
               (conj acc (str k ": " (get-or-throw val-map k))))) [] ks)
         (str/join "\n"))))


(defn verify-signature-header*
  [req sig-header-str]
  (let [{:keys [request-method uri headers server-name]} req
        sig-map     (->> (str/split sig-header-str #",")
                         (map #(-> %
                                   (str/replace #"\"" "")
                                   (str/split #"=")))
                         (into {}))
        sig-parts   (str/split (get sig-map "headers" "") #" ")
        sign-string (generate-signing-string (assoc headers "method" request-method
                                                            "path" uri
                                                            "host" server-name)
                                             sig-parts)
        signature   (get sig-map "signature")
        authority   (crypto/account-id-from-message sign-string signature)
        keyId       (get sig-map "keyId")
        auth        (if (= "na" keyId) nil keyId)
        _           (log/debug "Verifying signature. Sign string: " sign-string " Signature: " signature " Account Id: " auth)]
    {:auth      (or auth authority)
     :authority authority
     :type      :http-signature}))


(defn verify-signature-header
  "Verifies signed http request. If signature header does not exist,
  returns nil. If it exists and is valid, returns the authid associated
  with it. If it exists and is invalid, throws exception."
  [req]
  (let [sig-header-str (get (:headers req) "signature")]
    (when sig-header-str
      (verify-signature-header* req sig-header-str))))


(defn verify-digest
  "If a message digest is present, verifies it."
  [req]
  (if-let [digest (get-in req [:headers "digest"])]
    (let [[_ hash-type hash] (re-find #"^([^=]+)=(.+)$" digest)
          _           (when-not (#{"SHA-256"} hash-type)
                        (throw (ex-info (str "Digest type of " hash-type " is not supported.")
                                        {:status 401 :error :db/invalid-auth})))
          body        (:body req)
          body        (if (string? body) body (json/stringify body))
          calc-digest (case hash-type
                        "SHA-256" (crypto/sha2-256 body :base64))
          valid?      (= hash calc-digest)]
      (if-not valid?
        (throw (ex-info (str "Invalid digest.")
                        {:status 401 :error :db/invalid-auth}))
        req))
    req))


(defn sign-request
  "Signs http request by creating required headers and using supplied private key.
  req-type should be :get, :post, etc."
  ([req-method url request private-key]
   (sign-request req-method url request private-key nil))
  ([req-method url request private-key auth]
   (let [{:keys [headers body]} request
         URL              (URL. url)
         path             (str/lower-case (.getPath URL))
         host             (str/lower-case (.getHost URL))
         date             (or (get headers "date")
                              (-> (DateTimeFormatter/RFC_1123_DATE_TIME) (.format (ZonedDateTime/now (ZoneOffset/UTC)))))
         digest           (when body (format "SHA-256=%s" (crypto/sha2-256 body :base64)))
         sign-headers     (if body
                            ["(request-target)" "host" "date" "digest"]
                            ["(request-target)" "host" "date"])
         sign-vals        (if body
                            {"host" host "date" date "digest" digest "method" req-method "path" path}
                            {"host" host "date" date "method" req-method "path" path})
         signing-string   (generate-signing-string sign-vals sign-headers)
         sign-headers-str (str/join " " sign-headers)
         sig              (crypto/sign-message signing-string private-key)
         auth             (or auth "na")
         sig-header       (format "keyId=\"%s\",headers=\"%s\",algorithm=\"ecdsa-sha256\",signature=\"%s\"" auth sign-headers-str sig)]
     (assoc request :headers
                    (merge headers {"date"      (or (get headers "date")
                                                    (-> (DateTimeFormatter/RFC_1123_DATE_TIME) (.format (ZonedDateTime/now (ZoneOffset/UTC)))))
                                    "digest"    digest
                                    "signature" sig-header})))))


(defn verify-request*
  "Returns auth record from separated request parts."
  ([req method action db-name host]
   (verify-request* req method (str "/fdb/" (util/keyword->str db-name) "/" (util/keyword->str action))
                    host))
  ([req method uri host]
   (-> req
       (verify-digest)
       (assoc :request-method method
              :uri uri
              :server-name host)
       (verify-signature-header))))


(defn verify-request
  "Returns map of auth and authority from request."
  [request]
  (-> request
      (verify-digest)
      (verify-signature-header)))



(comment

  (def sig-str (generate-signing-string {"date"   "Tue, 11 Mar 2019 19:20:12 GMT"
                                         "method" :post
                                         "path"   "/fdb/test/one/query"
                                         "host"   "localhost"
                                         "digest" "SHA-256=7H1ZpcmGuWumeVlRzVWnfrYLfDyamUg5Y8Km49rb/c8="}
                                        ["(request-target)" "host" "date" "digest"]))

  (def myrequest {:headers {"content-type" "application/json"}
                  :body    (fluree.db.util.json/stringify {:select ["*"] :from "_collection"})})

  (def signed-request
    (sign-request :post "http://localhost:8090/fdb/test/permissions/query" myrequest "78f2ee93ef8008a270ffad949799462474f44c1ee8b29f07ec4fe1ae965b92c"))


  signed-request

  (verify-request* signed-request :query "test/permissions" "localhost")

  )
