(ns fluree.db.query.sparql-parser
  (:require #?(:clj  [instaparse.core :as insta]
               :cljs [instaparse.core :as insta :refer-macros [defparser]])
            #?(:cljs [fluree.db.util.cljs-shim :as cljs-shim :refer-macros [inline-resource]])
            [clojure.string :as str]))

#?(:cljs (def inline-content (fluree.db.util.cljs-shim/inline-resource "sparql-js.bnf")))

#?(:clj  (def sparql (insta/parser (clojure.java.io/resource "sparql.bnf")))
   :cljs (defparser sparql inline-content))


(def wikidata-prefixes
  ["<http://www.wikidata.org/entity/>"
   "<http://www.w3.org/2002/07/owl#>"
   "<http://www.w3.org/ns/prov#>"
   "<http://www.bigdata.com/queryHints#>"
   "<http://www.wikidata.org/prop/novalue/>"
   "<http://www.wikidata.org/prop/statement/value/>"
   "<http://www.wikidata.org/prop/qualifier/value/>"
   "<http://www.w3.org/1999/02/22-rdf-syntax-ns#>"
   "<http://www.wikidata.org/prop/>"
   "<http://www.wikidata.org/reference/>"
   "<http://www.wikidata.org/prop/statement/>"
   "<http://www.bigdata.com/rdf/search#>"
   "<http://schema.org/>"
   "<http://www.wikidata.org/prop/direct/>"
   "<http://www.wikidata.org/value/>"
   "<http://www.wikidata.org/wiki/Special:EntityData/>"
   "<http://www.bigdata.com/rdf/gas#>"
   "<http://www.wikidata.org/entity/statement/>"
   "<http://www.bigdata.com/rdf#>"
   "<http://www.wikidata.org/prop/qualifier/value-normalized/>"
   "<http://wikiba.se/ontology#>"
   "<http://www.wikidata.org/prop/reference/value-normalized/>"
   "<http://www.w3.org/2000/01/rdf-schema#>"
   "<http://www.wikidata.org/prop/qualifier/>"
   "<http://www.wikidata.org/prop/statement/value-normalized/>"
   "<http://www.wikidata.org/prop/reference/value/>"
   "<http://www.w3.org/2004/02/skos/core#>"
   "<http://www.wikidata.org/prop/reference/>"
   "<http://www.w3.org/2001/XMLSchema#>"])

(defn get-select
  [sparql-parsed]
  (let [query  (if (= :Query (first sparql-parsed))
                 (rest sparql-parsed))
        select (if (= :SelectQuery (first (second query)))
                 (rest (second query))
                 (throw (ex-info "Currently only SELECT queries are supported. Query type provided: " (first (second query))
                                 {:status 400
                                  :error  :db/invalid-query})))] select))

(defn var-from-var-parsed
  [var-vector]
  (let [rest-select (second var-vector)
        var'        (if (= :VAR1 (first rest-select))
                      (rest rest-select))

        var''       (if (= "?" (first var'))
                      (if (= :VARNAME (first (second var')))
                        (str "?" (str/join "" (rest (second var'))))
                        nil))
        var*        (if var'' var'' rest-select)]
    var*))

(defn prefixed-name-to-var
  [prefixed-name]
  (let [elements-of-str  (rest prefixed-name)
        elements-of-str' (if (and (vector? (last elements-of-str))
                                  (= :VARNAME (first (last elements-of-str))))
                           (concat (drop-last elements-of-str)
                                   (rest (last elements-of-str)))
                           elements-of-str)]
    (str/join "" elements-of-str')))

(defn var-or-ref-to-var
  [var-or-ref]
  (cond (= :Var (first var-or-ref))
        (var-from-var-parsed var-or-ref)

        (= :IRIref (first var-or-ref))
        (if (= :PrefixedName (first (second var-or-ref)))
          (prefixed-name-to-var (second var-or-ref)))

        (= :GraphTerm (first var-or-ref))
        (cond
          (= :NumericLiteral (first (second var-or-ref)))
          (#?(:clj read-string :cljs cljs.reader/read-string) (second (second var-or-ref)))

          (= :IRIref (first (second var-or-ref)))
          (if (= :PrefixedName (first (second (second var-or-ref))))
            (prefixed-name-to-var (second (second var-or-ref))))

          (= :RDFLiteral (first (second var-or-ref)))
          (if (= :String (first (second (second var-or-ref))))
            (let [elements-of-str  (rest (second (second var-or-ref)))
                  elements-of-str' (if (= "\"" (last elements-of-str))
                                     (drop-last elements-of-str) elements-of-str)] (str/join "" elements-of-str'))))))



(defn get-select-vars
  [sparql-parsed]
  (let [select       (get-select sparql-parsed)
        select-vars  (reduce
                       (fn [acc select-item]
                         (cond
                           (vector? select-item)
                           (cond
                             (= :Var (first select-item))
                             (conj acc (var-or-ref-to-var select-item))
                             (= :AggregateVar (first select-item))
                             (let [rest-select      (rest select-item)
                                   aggregate-clause (reduce (fn [acc select-item]
                                                              (if (vector? select-item)
                                                                (if (= :Var (first select-item))
                                                                  (conj acc (str/join "" (var-or-ref-to-var select-item)))
                                                                  (conj acc select-item))
                                                                (conj acc select-item))) [] rest-select)
                                   [aggregate-function function-param] (if (and (= "DISTINCT" (second aggregate-clause)) (= "count" (str/lower-case (first aggregate-clause))))
                                                                         ["count-distinct" (nth aggregate-clause 2)]
                                                                         [(str/lower-case (first aggregate-clause)) (str/join " " (rest aggregate-clause))])]
                               (conj acc (str "(" aggregate-function " " function-param " " ")")))

                             :else
                             acc)

                           (= "*" select-item)
                           (conj acc select-item)

                           :else acc)) [] select)
        select-vars' (into [] (distinct select-vars))]
    select-vars'))


(defn get-all-objects
  [object-list]
  (reduce (fn [acc object]
            (if (and (vector? object)
                     (= :Object (first object)))
              (if (= :VarOrTerm (first (second object)))
                (conj acc (var-or-ref-to-var (second (second object))))
                acc)
              acc))
          [] object-list))

(defn triples-same-subject-to-triples
  [triple-same-subject]
  (let [e      (if (= :VarOrTerm (first (first triple-same-subject)))
                 (var-or-ref-to-var (second (first triple-same-subject))))
        e-a-vs (if (= 2 (count triple-same-subject))
                 (reduce-kv
                   (fn [acc key subject]
                     (cond (= :PropertyListNotEmpty subject)
                           acc

                           (or (= :Var (first subject)) (= :IRIref (first subject)))
                           (let [a     (var-or-ref-to-var subject)
                                 v-vec (get-all-objects (nth (second triple-same-subject) (+ 1 key)))]
                             (reduce (fn [acc-vec v-value]
                                       (conj acc-vec [e a v-value]))
                                     acc v-vec))
                           :else acc))
                   [] (second triple-same-subject)))]
    e-a-vs))

(defn triples-block-to-triples
  ([triples-block]
   (triples-block-to-triples [] triples-block))
  ([triples triples-block]
   (if (= :TriplesSameSubject (first triples-block))
     (concat triples (triples-same-subject-to-triples (rest triples-block)))
     (if (= :TriplesBlock (first triples-block))
       (concat triples (reduce triples-block-to-triples triples (rest triples-block)))
       triples))))

(defn get-group-graph-pattern
  [sparql-parsed]
  (let [select         (get-select sparql-parsed)
        where          (-> (filterv #(= :WhereClause (first %)) select)
                           first rest first)
        triples-blocks (if (= :GroupGraphPattern (first where))
                         (rest where))]
    triples-blocks))

(defn get-triples
  [sparql-parsed]
  (let [triples-blocks (get-group-graph-pattern sparql-parsed)
        triples        (reduce (fn [vec triple]
                                 (if (and (vector? triple) (#{:TriplesSameSubject :TriplesBlock} (first triple)))
                                   (concat vec (triples-block-to-triples triple))
                                   vec)) [] triples-blocks)
        triples'       (into [] (distinct triples))]
    triples'))

(defn get-service-triples
  [sparql-parsed]
  (let [triples-blocks (get-group-graph-pattern sparql-parsed)
        triples        (reduce (fn [vec triple]
                                 (if (and (vector? triple) (#{:Service} (first triple)))
                                   (concat vec (triples-block-to-triples (second triple)))
                                   vec)) [] triples-blocks)]
    triples))


(defn get-optional-triples
  [sparql-parsed]
  (let [triples-blocks (get-group-graph-pattern sparql-parsed)
        triples        (reduce (fn [vec triple]
                                 (if (and (vector? triple) (#{:OptionalGraphPattern} (first triple)))
                                   (concat vec (triples-block-to-triples (second (second triple))))
                                   vec)) [] triples-blocks)]
    triples))


(defn get-language
  [service-triples]
  (-> (filterv #(and (= (first %) "bd:serviceParam")
                     (= (second %) "wikibase:language"))
               service-triples)
      first
      (nth 2)))

(defn get-solution-modifiers
  [sparql-parsed]
  (let [select             (get-select sparql-parsed)
        solution-modifiers (if (= :SolutionModifier (first (last select)))
                             (rest (last select)))
        modifiers          (reduce (fn [acc modifier]
                                     (cond

                                       (= :LimitClause (first modifier))
                                       (assoc acc :limit (#?(:clj read-string :cljs cljs.reader/read-string) (second modifier)))

                                       (= :OffsetClause (first modifier))
                                       (assoc acc :offset (#?(:clj read-string :cljs cljs.reader/read-string) (second modifier)))

                                       (= :OrderClause (first modifier))
                                       (cond
                                         (= :Var (first (second modifier)))
                                         (assoc acc :orderBy (var-or-ref-to-var (second modifier)))

                                         (#{"ASC" "DESC"} (second modifier))
                                         (assoc acc :orderBy [(second modifier) (var-or-ref-to-var (second (second (nth modifier 2))))])
                                         :else
                                         acc)

                                       :else
                                       acc)) {} solution-modifiers)] modifiers))

(defn remove-prefix-from-triple
  [triple]
  "Mainly used for removing fdb prefix from a triple"
  (map (fn [element]
         (if-not
           (or (number? element)
               (map? element)
               (nil? #?(:cljs (re-find #"[a-zA-Z0-9]+:_?[a-zA-Z0-9]*/?[a-zA-Z0-9]*" element)
                        :clj  (re-find (re-matcher #"[a-zA-Z0-9]+:_?[a-zA-Z0-9]*\/?[a-zA-Z0-9]*" element)))))
           (subs element (inc (str/index-of element ":")))
           element))
       triple))

(defn remove-wd-prefix-from-triple
  [triple]
  "Removing wd prefix from a triple if wd:?variable in triple-
  this is used to specify source when all elements in a triple are variables."
  (map (fn [element]
         (if-not
           (or (number? element)
               (map? element)
               (nil? #?(:cljs (re-find #"wd:\?.*" element)
                        :clj  (re-find (re-matcher #"wd:\?.*" element)))))
           (subs element (inc (str/index-of element ":")))
           element)) triple))


(defn triple-to-four-tuple
  "Takes a prefixed triple and turns into a four-tuple with source as the first element in the tuple."
  ([raw-triple]
   (triple-to-four-tuple raw-triple {}))
  ([raw-triple prefix-map]
   (let [source         (some (fn [element] (if-not (or (number? element) (map? element)
                                                        (nil? #?(:cljs (re-find #"[a-zA-Z0-9]+:_?[a-zA-Z0-9]*/?[a-zA-Z0-9]*" element)
                                                                 :clj  (re-find (re-matcher #"[a-zA-Z0-9]+:_?[a-zA-Z0-9]*\/?[a-zA-Z0-9]*" element)))))
                                              (subs element 0 (str/index-of element ":")))) raw-triple)
         source'        (cond (str/includes? source "wd") "$wd"
                              (some #(= (get prefix-map source) %) wikidata-prefixes) "$wd"
                              (str/includes? source "fdb") (str "$" source)
                              :else "$fdb")
         cleaned-triple (if (str/includes? source' "$fdb")
                          (remove-prefix-from-triple raw-triple)
                          (remove-wd-prefix-from-triple raw-triple))]
     (into [] (concat [source'] cleaned-triple)))))


(defn get-prefixes
  [sparql-parsed]
  (let [query      (if (= :Query (first sparql-parsed))
                     (rest sparql-parsed))
        prefixes   (if (= :Prologue (first (first query)))
                     (rest (first query))
                     (throw (ex-info (str "Improperly formatted SPARQL query. Note: FlureeDB does not support all SPARQL features.")
                                     {:status 400
                                      :error  :db/invalid-query})))
        prefix-map (reduce (fn [acc prefix]
                             (let [ref         (if (and (vector? (last prefix))
                                                        (= :IRI_REF (first (last prefix))))
                                                 (second (last prefix)))
                                   prefix-abbr (if (and (= :PrefixDecl (first prefix)) ref)
                                                 (->> prefix
                                                      drop-last
                                                      rest
                                                      drop-last
                                                      (str/join "")))] (assoc acc prefix-abbr ref)))
                           {} prefixes)]
    prefix-map))

(defn distinct-query?
  [sparql-parsed]
  (or (some #(= "DISTINCT" %) (get-select sparql-parsed)) false))

(defn pretty-print?
  [sparql-parsed]
  (let [print? (if (< 3 (count sparql-parsed))
                 (nth sparql-parsed 3)
                 false)]
    (if print?
      (boolean (second print?)) false)))

(defn sparql-to-ad-hoc
  [sparql-query]
  (let [sparql-parsed (sparql sparql-query)
        _             (if (= instaparse.gll.Failure (type sparql-parsed))
                        (throw (ex-info (str "Improperly formatted SPARQL query. Note: FlureeDB does not support all SPARQL features. Provided: " sparql-query)
                                        {:status 400
                                         :error  :db/invalid-query})))
        select        (get-select-vars sparql-parsed)
        prefix-map    (get-prefixes sparql-parsed)
        distinct?     (distinct-query? sparql-parsed)
        selectKey     (if distinct? :selectDistinct :select)
        labelLang     (get-language (get-service-triples sparql-parsed))
        modifiers     (get-solution-modifiers sparql-parsed)
        prettyPrint   (pretty-print? sparql-parsed)
        where         (mapv #(triple-to-four-tuple % prefix-map) (get-triples sparql-parsed))
        optional      (mapv #(triple-to-four-tuple % prefix-map) (get-optional-triples sparql-parsed))]
    {selectKey     select
     :where        where
     :optional     optional
     :prefixes     prefix-map
     :limit        (:limit modifiers)
     :orderBy      (:orderBy modifiers)
     :offset       (:offset modifiers)
     :wikidataOpts {:distinct distinct? :language labelLang}
     :prettyPrint  prettyPrint}))


(comment
  (def basic-query "SELECT (SUM(?nums) AS ?sum)\nWHERE {\n    ?person     fd:person/handle    \"zsmith\";\n                fd:person/favNums    ?nums.\n}")

  (def wd-query "SELECT ?handle ?title ?narrative_locationLabel\nWHERE {\n  ?user     fdb:person/favMovies    ?movie.\n  ?movie    fdb:movie/title       ?title.\n  ?wdMovie  wd:?label             ?title;\n            wd:P840               ?narrative_location;\n            wdt:P31               wd:Q11424.\n  ?user     fdb:person/handle       ?handle.\n}")


  (sparql basic-query)
  (sparql-to-ad-hoc basic-query)
  (sparql-to-ad-hoc wd-query)

  )
